@@ -4,9 +4,11 @@ import logging
 
 from tlz import merge, valmap
 
+from dask.utils import key_split
+
 from distributed.core import coerce_to_address, connect
 from distributed.diagnostics.progress import AllProgress
-from distributed.utils import color_of, key_split
+from distributed.utils import color_of
 from distributed.worker import dumps_function
 
 logger = logging.getLogger(__name__)
